// Copyright (C) 2011 by TOMITA Hiroshi.
//     This program is free software: you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation, either version 3 of the License, or
//     (at your option) any later version.

//     This program is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.

//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.

// object management

var CObject = function(b,s,g,w) {
// constructor
    this.block = b;
    this.start = s;
    this.goal = g;
    this.waypoint = w;

    this.maxWPID = 0;
    this.maxpairID = 0;
    this.maxstartID = 0;

    this.links = Array();//[[1,0,1,2],[3,4,5,3]];// ID, goal, start, goal
    this.linking = false;
    this.linkFrom = {};
    this.linkTo = {};
    this.liveLinking = false;
    this.prevWPIndex = -1;

//link setup
    for(i=0; i<this.start.length; i++){
        if(this.start[i].startID > this.maxstartID)
            this.maxstartID = this.start[i].startID;
        if(this.start[i].goalID != -1)
            for(j=0; j<this.goal.length; j++)
                if(this.start[i].goalID == this.goal[j].pairID)
                    this.links[this.links.length] = [T_START,T_GOAL,i,j];
        if(this.start[i].firstWPID != -1)
            for(j=0; j<this.waypoint.length; j++)
                if(this.start[i].firstWPID == this.waypoint[j].WPID)
                    this.links[this.links.length] = [T_START,T_WAYPOINT,i,j];
    }
    for(i=0; i<this.goal.length; i++){
        if(this.goal[i].pairID > this.maxpairID)
            this.maxpairID = this.goal[i].pairID;
        if(this.goal[i].nextStartID != -1)
            for(j=0; j<this.start.length; j++)
                if(this.goal[i].nextStartID == this.start[j].startID)
                    this.links[this.links.length] = [T_GOAL,T_START,i,j];
    }
    for(i=0; i<this.waypoint.length; i++){
        if(this.waypoint[i].WPID > this.maxWPID)
            this.maxWPID = this.waypoint[i].WPID;
        if(this.waypoint[i].nextWPID != -1)
            for(j=0; j<this.waypoint.length; j++)
                if(this.waypoint[i].nextWPID == this.waypoint[j].WPID)
                    this.links[this.links.length] = [T_WAYPOINT,T_WAYPOINT,i,j];
    }


    this.validate = function(type, num){
        var newobj;
        switch(type){
        case T_BLOCK:
            newobj = this.block[num];
            if (countObjParam(newobj) != 6){
                message("ERROR - invalid:"+type+"."+num+" length");
                return false;
            }
            break;
        case T_START:
            newobj = this.start[num];
            if (countObjParam(newobj) != 12){
                message("ERROR - invalid:"+type+"."+num+" length");
                return false;
            }
            break;
        case T_GOAL:
            newobj = this.goal[num];
            if (countObjParam(newobj) != 9){
                message("ERROR - invalid:"+type+"."+num+" length");
                return false;
            }
            break;
        case T_WAYPOINT:
            newobj = this.waypoint[num];
            if (countObjParam(newobj) != 5){
                message("ERROR - invalid:"+type+"."+num+" length");
                return false;
            } else {
                return true;
            }
            break;
        }
        if (newobj.shape == S_STROKERECT){
            message("ERROR - invalid: "+type+"." + num + " S_STROKERECT "); // debug only
            return false;
        }
        if (newobj.width == 0 || newobj.height == 0){
            message("ERROR - invalid: "+type+"." + num + " width or height "); // debug only
            return false;
        }

        if (!isInCanvas(newobj.x, newobj.y))
            return false;
        
        if (newobj.shape == S_CIRCLE){
            newobj.height = 360; newobj.degree = 0;
        }
        if (newobj.width < 0){
            newobj.width = -newobj.width;
        }
        if (newobj.height < 0){
            if (newobj.shape == S_CIRCLE || newobj.shape == S_STROKECIRCLE){
                newobj.height = newobj.height%360+360;
            } else if (newobj.shape == S_RECT){
                newobj.height = -newobj.height;
            }
        }
        newobj.degree = (newobj.degree%360 + 360)%360;// to plus
        if ( newobj.shape == S_RECT) {
            newobj.degree = (newobj.degree%180+180)%180;
        }

        return true;
    };



    this.strictValidate = function(){
        this.links = new Array; // delete links
        for(it=0;it<this.block.length; it++){
            if (isEmpty(this.block[it])){
                this.block.splice(it,1);
                it --;
                continue;
            }
            
            if (!this.validate(T_BLOCK,it)){
                this.remove(T_BLOCK, it, false);
                this.block.splice(it,1);
                it --;
                continue;
            }
        }

        for(it=0; it<this.start.length; it++){
            if (isEmpty(this.start[it])){
                this.start.splice(it,1);
                it --;
                continue;
            }

            if (!this.validate(T_START,it)){
                this.remove(T_START, it, false);
                this.start.splice(it,1);
                it --;
                continue;
            }
            if (this.start[it].goalID == -1){
                this.remove(T_START,it, false);
                this.start.splice(it,1);
                it --;
                continue;
            } else { // is exist?
                for(j=0; j<this.goal.length; j++){
                    if(this.start[it].goalID == this.goal[j].pairID)
                        break;
                }
                if (j == this.goal.length){ // not hit
                    this.remove(T_START, it, false);
                    this.start.splice(it,1);
                    it --;
                    continue;
                }
            }
            if (this.start[it].firstWPID == -1){
                if(this.start[it].goalID == this.goal[j].pairID)
                    this.start[it].firstWPID = this.goal[j].myWPID;
            } else {
                for (k=0; k<this.waypoint.length; k++){
                    if(this.waypoint[k].WPID == this.start[it].firstWPID)
                        break;
                }
                if (k == this.waypoint.length){
                    this.start[it].firstWPID = this.goal[j].myWPID;
                }
            }
            
            if(this.start[it].generateSpeed == 0){
                for(j = 0; j<this.goal.length; j++){
                    if(this.goal[j].nextStartID == this.start[it].startID){
                        break;
                        
                    }
                }
                if(j == this.goal.length){
                    this.remove(T_START,it);
                    this.start.splice(it,1);
                    it --;
                    continue;
                }
            }
        }
        
        for(it=0; it<this.goal.length; it++){
            linking = false;
            linked = false;
            if (isEmpty(this.goal[it])){
                this.goal.splice(it,1);
                it --;
                continue;
            }
            
            if (!this.validate(T_GOAL,it)){
                this.remove(T_GOAL,it, false);
                this.goal.splice(it,1);
                it --;
                continue;
            }
            for (j=0; j<this.start.length; j++){
                if (this.start[j].startID == this.goal[it].nextStartID)
                    linking = true;
                if (this.start[j].goalID == this.goal[it].pairID)
                    linked = true;
            }
            if (!linked){
                for (j=0; j<agent.agt.length; j++){
                    if (agent.agt[j][2] == this.goal[it].pairID)
                        linked = true;
                }
            }
            if ( !linked ){ // not linked from any agents
                this.remove(T_GOAL,it, false);
                this.goal.splice(it,1);
                it --;
                continue;
            }
            if ( this.goal[it].nextStartID == -1 && !linking ){ // not hit
                this.goal[it].nextStartID = -1; // disable
            }
        }

        for(it=0; it<this.waypoint.length; it++){
            if (isEmpty(this.waypoint[it])){
                this.waypoint.splice(it,1);
                it --;
                continue;
            }

            if (!this.validate(T_WAYPOINT,it)){
                this.remove(T_WAYPOINT,it,false);
                this.waypoint.splice(it,1);
                it --;
                continue;
            }
            if (this.waypoint[it].nextWPID == -1){ // if not goal.mywpid, delete
                for (j=0; j<this.goal.length; j++){
                    if(this.waypoint[it].WPID == this.goal[j].myWPID)
                        break;
                }
                if (j == this.goal.length){
                    this.remove(T_WAYPOINT,it,false);
                    this.waypoint.splice(it,1);
                    it --;
                    continue;
                }
            } else {
                for (j=0; j<this.waypoint.length; j++){
                    if (this.waypoint[j].WPID == this.waypoint[it].nextWPID)
                        break;
                }
                if (j == this.waypoint.length){
                    this.remove(T_WAYPOINT,it,false);
                    this.waypoint.splice(it,1);
                    it --;
                    continue;
                }
            }
        }

        for(i=0; i<this.start.length; i++){
            if(this.start[i].goalID != -1)
                for(j=0; j<this.goal.length; j++)
                    if(this.start[i].goalID == this.goal[j].pairID)
                        this.links[this.links.length] = [T_START,T_GOAL,i,j];
            if(this.start[i].firstWPID != -1)
                for(j=0; j<this.waypoint.length; j++)
                    if(this.start[i].firstWPID == this.waypoint[j].WPID)
                        this.links[this.links.length] = [T_START,T_WAYPOINT,i,j];
        }
        for(i=0; i<this.goal.length; i++){
            if(this.goal[i].nextStartID != -1)
                for(j=0; j<this.start.length; j++)
                    if(this.goal[i].nextStartID == this.start[j].startID)
                        this.links[this.links.length] = [T_GOAL,T_START,i,j];
        }
        for(i=0; i<this.waypoint.length; i++){
            if(this.waypoint[i].nextWPID != -1)
                for(j=0; j<this.waypoint.length; j++)
                    if(this.waypoint[i].nextWPID == this.waypoint[j].WPID)
                        this.links[this.links.length] = [T_WAYPOINT,T_WAYPOINT,i,j];
        }
    }



    this.add = function(){
        var backup = new Array;
        if(arguments.length == 2){
            type = arguments[0];
            params = arguments[1];
            safe_params = {};
            switch(type){
            case T_BLOCK:
                safe_params.x = ("x" in params) ? params.x : false ;
                safe_params.y = ("y" in params) ? params.y : false ;
                safe_params.width = ("width" in params) ? params.width : false;
                safe_params.height = ("height" in params) ? params.height : false;
                safe_params.shape = ("shape" in params) ? params.shape : false;

                for(i=0; i<countObjParam(safe_params); i++){
                    if(safe_params[i] === false) return false;
                }

                safe_params.degree = "degree" in params ? params.degree : 0;
                this.block[this.block.length] = safe_params;
                if(!this.validate(type,this.block.length-1)){
                    this.block.splice(this.block.length-1,1);
                    return false;
                } else {
                    return true;
                }
                break;

            case T_START:
                safe_params.x = ("x" in params) ? params.x : false ;
                safe_params.y = ("y" in params) ? params.y : false ;
                safe_params.width = ("width" in params) ? params.width : false;
                safe_params.height = ("height" in params) ? params.height : false;
                safe_params.shape = ("shape" in params) ? params.shape : false;

                for(i=0; i<countObjParam(safe_params); i++){
                    if(safe_params[i] === false) return false;
                }

                safe_params.degree = "degree" in params ? params.degree : 0;
                safe_params.generateSpeed = "generateSpeed" in params ? params.generateSpeed : 0;
                safe_params.goalID = "goalID" in params ? params.goalID : -1;
                safe_params.firstWPID = "firstWPID" in params ? params.firstWPID : -1;
                safe_params.startID = "startID" in params ? params.startID : ++this.maxstartID;
                safe_params.velocityMean = "velocityMean" in params ? params.velocityMean : 1.5;
                safe_params.velocitySD = "velocitySD" in params ? params.velocitySD : 0;
                this.start[this.start.length] = safe_params;
                
                if(!this.validate(type,this.start.length-1)){
                    this.start.splice(this.start.length-1,1);
                    return false;
                } else {
                    return true;
                }
                break;

            case T_GOAL:
                safe_params.x = ("x" in params) ? params.x : false ;
                safe_params.y = ("y" in params) ? params.y : false ;
                safe_params.width = ("width" in params) ? params.width : false;
                safe_params.height = ("height" in params) ? params.height : false;
                safe_params.shape = ("shape" in params) ? params.shape : false;

                for(i=0; i<countObjParam(safe_params); i++){
                    if(safe_params[i] === false) return false;
                }

                safe_params.degree = "degree" in params ? params.degree : 0;
                safe_params.nextStartID = "nextStartID" in params ? params.nextStartID : -1;
                safe_params.myWPID = "myWPID" in params ? params.myWPID : -1; // TODO : AUTOMATIC WPGENERATION
                safe_params.pairID = "pairID" in params ? params.pairID : ++this.maxpairID;
                this.goal[this.goal.length] = safe_params;
                if(!this.validate(type,this.goal.length-1)){
                    this.goal.splice(this.goal.length-1,1);
                    return false;
                } else {
                    return true;
                }
                break;
            case T_WAYPOINT:
                safe_params.x = ("x" in params) ? params.x : false ;
                safe_params.y = ("y" in params) ? params.y : false ;
                for(i=0; i<countObjParam(safe_params); i++){
                    if(safe_params[i] === false) return false;
                }
                safe_params.WPID = ++this.maxWPID;
                safe_params.nextWPID = -1;
                this.waypoint[this.waypoint.length] = safe_params;
                if(!this.validate(type,this.waypoint.length-1)){
                    this.waypoint.splice(this.waypoint.length-1,1);
                    return false;
                }
                if(this.liveLinking){
                    if(this.prevWPIndex != -1)
                        this.makeLink(T_WAYPOINT,T_WAYPOINT,this.prevWPIndex,this.waypoint.length-1);
                    this.prevWPIndex = this.waypoint.length-1;
                }
                break;
            }
            return true;
        } else {
            var shape,type,x,y,width,height;
            shape = arguments[0];
            type  = arguments[1];
            x     = arguments[2];
            y     = arguments[3];
            width = arguments[4];
            height= arguments[5];

            switch(type){
            case T_BLOCK:
                if (shape == S_STROKECIRCLE){
                    height = 360;
                }
                if(shape == S_STROKERECT){
                    this.add(S_RECT,type,x-width/2,y,2,height);
                    this.add(S_RECT,type,x,y-height/2,width,2);
                    this.add(S_RECT,type,x+width/2,y,2,height);
                    this.add(S_RECT,type,x,y+height/2,width,2);
                } else {
                    this.block[this.block.length] = {
                        x : x,
                        y : y,
                        width : width,
                        height : height,
                        shape : shape,
                        degree : 0
                    };
                    if(!this.validate(type,this.block.length-1)){
                        this.block.splice(this.block.length-1,1);
                        return false;
                    } else {
                        return true;
                    }
                }
                break;

            case T_START:
                for (i=0;i<this.start.length; i++){
                    backup[i] = this.start[i];
                }
                if (shape == S_STROKECIRCLE || shape == S_STROKERECT){
                    return false;
                }
                this.start[this.start.length] = {
                    x : x,
                    y : y,
                    width : width,
                    height : height,
                    degree : 0,
                    shape : shape,
                    startID : ++this.maxstartID,
                    goalID : -1,
                    firstWPID : -1,
                    generateSpeed : 0,
                    velocityMean: 1.5,
                    velocitySD: 0
                };
                if(!this.validate(type,this.start.length-1)){
                    this.start.splice(this.start.length-1,1);
                    return false;
                } else {
                    return true;
                }
                break;

            case T_GOAL:
                for (i=0;i<this.goal.length; i++){
                    backup[i] = this.goal[i];
                }
                if (shape == S_STROKECIRCLE || shape == S_STROKERECT){
                    return false;
                }
                this.goal[this.goal.length] = {
                    x : x,
                    y : y,
                    width : width,
                    height : height,
                    degree : 0,
                    shape : shape,
                    pairID : ++this.maxpairID,
                    nextStartID : -1,
                    myWPID : this.addWP(x,y,true)
                };
                if(!this.validate(type,this.goal.length-1)){
                    this.waypoint.splice(this.waypoint.length-1,1);
                    this.goal.splice(this.goal.length-1,1);
                    return false;
                } else {
                    return true;
                }
                break;
            case T_WAYPOINT:
                this.addWP(x,y,false);
                break;
            }
            return true;
        }
    };

    
    this.addWP  =  function(x, y , p){
        this.waypoint[this.waypoint.length] = {
            x : x,
            y : y,
            WPID : ++this.maxWPID,
            nextWPID : -1,
            p : p
        }
        if(this.liveLinking){
            if(this.prevWPIndex != -1)
                this.makeLink(T_WAYPOINT,T_WAYPOINT,this.prevWPIndex,this.waypoint.length-1);
            this.prevWPIndex = this.waypoint.length-1;
        }
        return this.waypoint[this.waypoint.length - 1].WPID;
    }
    


    this.changeParameter = function(id, val){
        var index = Number(id.substr(2,id.length-2));
        var data = Number(val);
        var type = 0;

        switch(id.substr(0,1)){
        case 'b':
            backup = this.block[index];
            type = T_BLOCK;
            switch(id.substr(1,1)){
            case 'x':
                this.block[index].x = data;
                break;
            case 'y':
                this.block[index].y = data;
                break;
            case 'w':
                this.block[index].width = data;
                break;
            case 'h':
                this.block[index].height = data;
                break;
            case 'd':
                this.block[index].degree = data;
                break;
            case 'r':
                break;
            }
            if (!this.validate(type,index)){
                this.block[index] = backup;
            }
            break;
        case 's':
            type = T_START;
            backup = this.start[index];
            switch(id.substr(1,1)){
            case 'x':
                this.start[index].x = data;
                break;
            case 'y':
                this.start[index].y = data;
                break;
            case 'w':
                this.start[index].width = data;
                break;
            case 'h':
                this.start[index].height = data;
                break;
            case 'd':
                this.start[index].degree = data;
                break;
            case 'r':
                this.start[index].generateSpeed = data;
                break;
            case 'm':
                this.start[index].velocityMean = data;
                break;
            case 's':
                this.start[index].velocitySD = data;
                break;
            }
            if (!this.validate(type,index)){
                this.start[index] = backup;
            }
            break;
        case 'g':
            type = T_GOAL;
            backup = this.goal[index];
            switch(id.substr(1,1)){
            case 'x':
                this.goal[index].x = data;
                break;
            case 'y':
                this.goal[index].y = data;
                break;
            case 'w':
                this.goal[index].width = data;
                break;
            case 'h':
                this.goal[index].height = data;
                break;
            case 'd':
                this.goal[index].degree = data;
                break;
            }
            if (!this.validate(type,index)){
                this.goal[index] = backup;
            }
            break;
        case 'w':
            type = T_WAYPOINT;
            backup=this.waypoint[index];
            if(this.waypoint[index].p) return false;
            switch(id.substr(1,1)){
            case 'x':
                this.waypoint[index].x = data;
                break;
            case 'y':
                this.waypoint[index].y = data;
                break;
            }
            if (!this.validate(type,index)){
                this.waypoint[index] = backup;
            }
            break;
        }
        revertAll();
    };


    this.remove = function(type, num){
        var obj = Array();
        if (arguments.length >= 3){
            for(i=num; i<num+arguments[2]; i++){
                this.remove(type,i,false);
            }
            return true;
        }
        switch(type){
        case T_BLOCK:
            obj = this.block;
            break;
        case T_START:
            obj = this.start;
            if(obj[num].goalID != -1){ // if has link
                this.cleanupLink(T_START, T_GOAL, num);
            }
            if(obj[num].firstWPID != -1){ // if has link
                this.cleanupLink(T_START, T_WAYPOINT, num);
            }
            for(i=0; i<this.goal.length; i++){
                if (this.goal[i].nextStartID == obj[num].startID)
                    this.cleanupLink(T_GOAL, T_START, i);
            }
            break;
        case T_GOAL:
        obj = this.goal;
        if(obj[num].nextStartID != -1){ // if has link
                this.cleanupLink(T_GOAL, T_START, num);
            }
            for(i=0; i<this.start.length; i++){
                if (this.start[i].goalID == obj[num].pairID)
                    this.cleanupLink(T_START, T_GOAL, i);
            }
            for(i=0; i<agent.agt.length; i++){
                if (agent.agt[i][2] == obj[num].pairID)
                    this.cleanupLink(T_PED, T_GOAL, i);
            }
        for(i=0;i<this.waypoint.length;i++){
            if(obj[num].myWPID == this.waypoint[i].WPID){
                this.remove(T_WAYPOINT,i);
                break;
            }
        }
            break;
        case T_WAYPOINT:
            obj = this.waypoint;
            if(obj[num].nextWPID != -1){ // if has link
                this.cleanupLink(T_WAYPOINT, T_WAYPOINT, num);
            }
            for(i=0; i<this.waypoint.length; i++){
                if (this.waypoint[i].nextWPID == obj[num].WPID)
                    this.cleanupLink(T_WAYPOINT, T_WAYPOINT, i);
            }
            for(i=0; i<this.start.length; i++){
                if (this.start[i].firstWPID == obj[num].WPID)
                    this.cleanupLink(T_START, T_WAYPOINT, i);
            }
            for(i=0; i<agent.agt.length; i++){
                if (agent.agt[i][3] == obj[num].WPID)
                    this.cleanupLink(T_PED, T_WAYPOINT, i);
            }            
            break;
        }
        obj[num] = {};
    };


    this.cleanupLink = function(from,to,num){
        for(i=0; i<this.links.length; i++){
            if (this.links[i][0] == from && this.links[i][2] == num &&
                this.links[i][1] == to){
                this.deleteLink(i);
                break;
            }
        }
    };


    this.toggleLiveLinking = function(){
        linking = false;
        this.prevWPIndex = -1;
        this.liveLinking = !this.liveLinking;
        if(this.liveLinking){
            $('#wp_linking').attr("checked","checked");
        } else {
            $('#wp_linking').removeAttr("checked");
        }
    }


    this.selectLinkObject = function(type, num){
        // get radio button status
        switch( $('input[name="tp_link_selector"]:checked').val() ){
        case 'from':
            $('#tp_link_sel_to').attr('checked','checked');
            if(isEmpty(this.linkTo)){
                this.linkFrom = [type,num];
                $('#tp_link_from_mark').html(_getIMG(type));
                $('#tp_link_from_number').text(num);
            } else {
                this.makeLink(type,this.linkTo[0],num,this.linkTo[1]);

                // reset forms
                $('#tp_link_from_mark').text('');
                $('#tp_link_from_number').text('');
                $('#tp_link_to_mark').text('');
                $('#tp_link_to_number').text('');
                this.linkFrom = {};
                this.linkTo = {};
                revertAll();
            }
            break;
        case 'to':
            $('#tp_link_sel_from').attr('checked','checked');
            if(isEmpty(this.linkFrom)){
                this.linkTo = [type,num];
                $('#tp_link_to_mark').html(_getIMG(type));
                $('#tp_link_to_number').text(num);
            } else {
                this.makeLink(this.linkFrom[0],type,this.linkFrom[1],num);

                // reset forms
                $('#tp_link_from_mark').text('');
                $('#tp_link_from_number').text('');
                $('#tp_link_to_mark').text('');
                $('#tp_link_to_number').text('');
                this.linkFrom = {};
                this.linkTo = {};
                revertAll();
            }
            break;
        }
    }


    this.mark = function(type, num){
        var obj = Object();
        if(arguments.length > 2)
            style = arguments[2];
        else
            style = null;

        if(mode == M_LINK) this.selectLinkObject(type, num);

        switch(type){
        case T_BLOCK:
            obj = this.block[num];
            break;
        case T_START:
            obj = this.start[num];
            break;
        case T_GOAL:
            obj = this.goal[num];
            break;
        case T_WAYPOINT:
            obj = this.waypoint[num];
            fillShape(S_CIRCLE, this.waypoint[num].x, this.waypoint[num].y, canvasRatio,360,0, WP_markupColor);
            return false;
            break;
        }
        
        switch(mode){
        case M_LINK:
            if(style === null) style = linkupColor;
            switch (obj.shape){
            case S_RECT:
                fillShape(S_STROKERECT, obj.x, obj.y, obj.width, obj.height, obj.degree, style);
                break;
            case S_CIRCLE:
            case S_STROKECIRCLE:
                fillShape(S_STROKECIRCLE,obj.x,obj.y,obj.width,obj.height,obj.degree,style);
                break;
            }
            break;
        case M_MAP:
            if(style === null) style = markupColor;
            switch (obj.shape){
            case S_RECT:
                fillShape(S_STROKERECT,obj.x,obj.y,obj.width,obj.height,obj.degree,style);
                break;
            case S_CIRCLE:
            case S_STROKECIRCLE:
                fillShape(S_STROKECIRCLE,obj.x,obj.y,obj.width,obj.height,obj.degree,style);
                break;
            }
            break;
        }
    };

    this.unmark = function(type, num){
        revertAll(false);
    };


    this.markLink = function(i){
        var fromx,fromy,tox,toy;
        switch(this.links[i][0]){
        case T_BLOCK:
            fromx = this.block[this.links[i][2]].x;
            fromy = this.block[this.links[i][2]].y;
            break;
        case T_START:
            fromx = this.start[this.links[i][2]].x;
            fromy = this.start[this.links[i][2]].y;
            break;
        case T_GOAL:
            fromx = this.goal[this.links[i][2]].x;
            fromy = this.goal[this.links[i][2]].y;
            break;
        case T_WAYPOINT:
            fromx = this.waypoint[this.links[i][2]].x;
            fromy = this.waypoint[this.links[i][2]].y;
            break;
        case T_PED:
            fromx = agent.agt[this.links[i][2]][0];
            fromy = agent.agt[this.links[i][2]][1];
        }
        switch(this.links[i][1]){
        case T_BLOCK:
            tox = this.block[this.links[i][3]].x;
            toy = this.block[this.links[i][3]].y;
            break;
        case T_START:
            tox = this.start[this.links[i][3]].x;
            toy = this.start[this.links[i][3]].y;
            break;
        case T_GOAL:
            tox = this.goal[this.links[i][3]].x;
            toy = this.goal[this.links[i][3]].y;
            break;
        case T_WAYPOINT:
            tox = this.waypoint[this.links[i][3]].x;
            toy = this.waypoint[this.links[i][3]].y;
            break;
        }
        if(arguments.length != 0){
            drawLine(fromx,fromy,tox,toy,markupColor);
        }
    }
    this.unmarkLink = function(i){
        revertAll();
    }



    this.showLinkLine = function(){
        var fromx,fromy,tox,toy;
        if(this.linking == true ){
            this.mark(this.linkFrom, linkFromColor);
        }
        for(i=0;i<this.links.length; i++){
            switch(this.links[i][0]){
            case T_BLOCK:
                fromx = this.block[this.links[i][2]].x;
                fromy = this.block[this.links[i][2]].y;
                break;
            case T_START:
                fromx = this.start[this.links[i][2]].x;
                fromy = this.start[this.links[i][2]].y;
                break;
            case T_GOAL:
                fromx = this.goal[this.links[i][2]].x;
                fromy = this.goal[this.links[i][2]].y;
                break;
            case T_WAYPOINT:
                fromx = this.waypoint[this.links[i][2]].x;
                fromy = this.waypoint[this.links[i][2]].y;
                break;
            case T_PED:
                fromx = agent.agt[this.links[i][2]][0];
                fromy = agent.agt[this.links[i][2]][1];
            }
            switch(this.links[i][1]){
            case T_BLOCK:
                tox = this.block[this.links[i][3]].x;
                toy = this.block[this.links[i][3]].y;
                break;
            case T_START:
                tox = this.start[this.links[i][3]].x;
                toy = this.start[this.links[i][3]].y;
                break;
            case T_GOAL:
                tox = this.goal[this.links[i][3]].x;
                toy = this.goal[this.links[i][3]].y;
                break;
            case T_WAYPOINT:
                tox = this.waypoint[this.links[i][3]].x;
                toy = this.waypoint[this.links[i][3]].y;
                break;
            }
            drawLine(fromx,fromy,tox,toy,linklineColor);
        }
    };

    this.makeLink = function (from_type, to_type, from_i, to_i){
        switch(from_type){
        case T_START:
            switch(to_type){
            case T_GOAL:
                if(this.start[from_i].goalID != -1){
                    this.cleanupLink(T_START, T_GOAL, from_i);
                }
                this.start[from_i].goalID = this.goal[to_i].pairID;
                if (this.start[from_i].firstWPID == -1)
                    this.start[from_i].firstWPID = this.goal[to_i].myWPID;
                break;
            case T_WAYPOINT:
                if(this.start[from_i].firstWPID != -1){
                    this.cleanupLink(T_START, T_WAYPOINT, from_i);
                }
                this.start[from_i].firstWPID = this.waypoint[to_i].WPID;
                break;
            default:
                return false;
            }
            break;
        case T_GOAL:
            switch(to_type){
            case T_START:
                if(this.goal[from_i].nextStartID != -1){
                    this.cleanupLink(T_GOAL, T_START, from_i);
                }
                this.goal[from_i].nextStartID = this.start[to_i].startID;
                break;
            default:
                return false;
            }
            break;
        case T_WAYPOINT:
            switch(to_type){
            case T_WAYPOINT:
                if(this.waypoint[from_i].nextWPID != -1){
                    this.cleanupLink(T_WAYPOINT, T_WAYPOINT, from_i);
                }
                this.waypoint[from_i].nextWPID = this.waypoint[to_i].WPID;
                break;
            default:
                return false;
            }   
            break;
        case T_PED:
            switch(to_type){
            case T_WAYPOINT:
                if(agent.agt[from_i][3] != -1){
                    this.cleanupLink(T_PED, T_WAYPOINT, from_i);
                }
                agent.agt[from_i][3] = this.waypoint[to_i].WPID;
                break;
            case T_GOAL:
                if(agent.agt[from_i][2] != -1){
                    this.cleanupLink(T_PED, T_GOAL, from_i);
                }
                agent.agt[from_i][2] = this.goal[to_i].pairID;
                if(agent.agt[from_i][3] == -1)
                    agent.agt[from_i][3] - this.goal[to_i].myWPID;
                break;
            default:
                return false;
            }
            break;
        }
        this.links[this.links.length] = [from_type,to_type,from_i,to_i];
        return true;
    };



    this.show = function(){
        for (i = 0; i<this.block.length; i++){
            if(isEmpty(this.block[i])) continue;
            fillShape(this.block[i].shape,this.block[i].x,this.block[i].y,this.block[i].width,this.block[i].height
                      ,this.block[i].degree,objColor[T_BLOCK]);
        }
        for (i = 0; i<this.start.length; i++){
            if(isEmpty(this.start[i])) continue;
            fillShape(this.start[i].shape,this.start[i].x,this.start[i].y,this.start[i].width,this.start[i].height
                      ,this.start[i].degree,objColor[T_START]);
        }
        for (i = 0; i<this.goal.length; i++){
            if(isEmpty(this.goal[i])) continue;
            fillShape(this.goal[i].shape,this.goal[i].x,this.goal[i].y,this.goal[i].width,this.goal[i].height
                      ,this.goal[i].degree,objColor[T_GOAL]);
        }
        for (i = 0; i<this.waypoint.length; i++){
            if(isEmpty(this.waypoint[i])) continue;
            fillShape(S_CIRCLE,this.waypoint[i].x,this.waypoint[i].y,canvasRatio,360
                      ,0,objColor[T_WAYPOINT]);
        }
    };


    this.list = function() {
        this.listMap();
        this.listLink();
    };


    this.listMap = function(){
        this._blockListMap();
        this._startListMap();
        this._goalListMap();
        this._wpListMap();        
    };


    this._blockListMap = function(){
        $table = $("#tab-blocks > table.container").text('');
        $table.append("<tr class=\"listhead\"><th>ID</th><th>shape</th>"+
                                "<th>x</th><th>y</th><th>width</th><th>height</th><th>deg</th>"+
                                "<th>del</th></tr>");
        for (i=0; i<this.block.length; i++){
            if(isEmpty(this.block[i])) continue;
            obj = this.block[i];
            s = '<tr class="object" id="object'+i+'" onmousedown="object.mark('+T_BLOCK+','+i+')" onmouseup="object.unmark('+T_BLOCK+','+i+')">';
            s += "<td>"+i+"</td>"+
                "<td>"+_getIMG(obj.shape)+"</td>"+
                "<td><input type='text' id='bx"+i+"' onchange='object.changeParameter(this.id ,$(this).val())'"+
                " class='objectParam' onfocus='this.select()' size='3' value='"+obj.x+"' /></td>"+
                "<td><input type='text' id='by"+i+"' onchange='object.changeParameter(this.id ,$(this).val())'"+
                " class='objectParam' onfocus='this.select()' size='3' value='"+obj.y+"' /></td>"+
                "<td><input type='text' id='bw"+i+"' onchange='object.changeParameter(this.id ,$(this).val())'"+
                " class='objectParam' onfocus='this.select()' size='3' value='"+obj.width+"' /></td>";
            if (obj.shape == S_CIRCLE){
                s += "<td></td><td></td>";
            } else {
                s +=
                "<td><input type='text' id='bh"+i+"' onchange='object.changeParameter(this.id ,$(this).val())'"+
                    " class='objectParam' onfocus='this.select()' size='3' value='"+obj.height+"' /></td>"+
                "<td><input type='text' id='bd"+i+"' onchange='object.changeParameter(this.id ,$(this).val())'"+
                    " class='objectParam' onfocus='this.select()' size='3' value='"+obj.degree+"' /></td>";
            }
            s += "<td><a class='tool' href='#' onmouseup='object.remove("+T_BLOCK+","+i+"); revertAll(); return false;'>[D]</a></td></tr>";
            $table.append(s);
        }
    };


    this._startListMap = function(){
        $table = $("#tab-starts > table.container").text('');
        $table.append("<tr class=\"listhead\"><th>ID</th><th>shape</th>"+
                                "<th>x</th><th>y</th><th>width</th><th>height</th><th>deg</th>"+
                                "<th>rate</th><th>vM</th><th>vSD</th>"+
                                "<th>del</th></tr>");
        for (i=0; i<this.start.length; i++){
            if(isEmpty(this.start[i])) continue;
            obj = this.start[i];
            s = obj.goalID<0 ? 
                '<tr class="object invalid" id="object'+i+'" onmousedown="object.mark('+T_START+','+i+')" onmouseup="object.unmark('+T_START+','+i+')">' :
                '<tr class="object" id="object'+i+'" onmousedown="object.mark('+T_START+','+i+')" onmouseup="object.unmark('+T_START+','+i+')">';
            s += "<td>"+i+"</td>"+
                "<td>"+_getIMG(obj.shape)+"</td>"+
                "<td><input type='text' id='sx"+i+"' onchange='object.changeParameter(this.id ,$(this).val())'"+
                " class='objectParam' onfocus='this.select()' size='3' value='"+obj.x+"' /></td>"+
                "<td><input type='text' id='sy"+i+"' onchange='object.changeParameter(this.id ,$(this).val())'"+
                " class='objectParam' onfocus='this.select()' size='3' value='"+obj.y+"' /></td>"+
                "<td><input type='text' id='sw"+i+"' onchange='object.changeParameter(this.id ,$(this).val())'"+
                " class='objectParam' onfocus='this.select()' size='3' value='"+obj.width+"' /></td>";
            if (obj.shape == S_CIRCLE){
                s += "<td></td><td></td>";
            } else {
                s +=
                "<td><input type='text' id='sh"+i+"' onchange='object.changeParameter(this.id ,$(this).val())'"+
                    " class='objectParam' onfocus='this.select()' size='3' value='"+obj.height+"' /></td>"+
                "<td><input type='text' id='sd"+i+"' onchange='object.changeParameter(this.id ,$(this).val())'"+
                    " class='objectParam' onfocus='this.select()' size='3' value='"+obj.degree+"' /></td>";
            }
            s += "<td><input type='text' id='sr"+i+"' onchange='object.changeParameter(this.id ,$(this).val())'"+
                    " class='objectParam' onfocus='this.select()' size='2' value='"+obj.generateSpeed+"' /></td>";
            s += "<td><input type='text' id='sm"+i+"' onchange='object.changeParameter(this.id ,$(this).val())'"+
                    " class='objectParam' onfocus='this.select()' size='2' value='"+obj.velocityMean+"' /></td>";
            s += "<td><input type='text' id='ss"+i+"' onchange='object.changeParameter(this.id ,$(this).val())'"+
                    " class='objectParam' onfocus='this.select()' size='2' value='"+obj.velocitySD+"' /></td>";
            s += "<td><a class='tool' href='#' onmouseup='object.remove("+T_START+","+i+"); revertAll(); return false;'>[D]</a></td></tr>";
            $table.append(s);
        }
    };

    this._goalListMap = function(){
        $table = $("#tab-goals > table.container").text('');
        $table.append("<tr class=\"listhead\"><th>ID</th><th>shape</th>"+
                                "<th>x</th><th>y</th><th>width</th><th>height</th><th>deg</th>"+
                                "<th>del</th></tr>");
        for (i=0; i<this.goal.length; i++){
            if(isEmpty(this.goal[i])) continue;
            obj = this.goal[i];
            s = '<tr class="object" id="object'+i+'" onmousedown="object.mark('+T_GOAL+','+i+')" onmouseup="object.unmark('+T_GOAL+','+i+')">';
            s += "<td>"+this.goal[i].pairID+"</td>"+
                "<td>"+_getIMG(obj.shape)+"</td>"+
                "<td><input type='text' id='gx"+i+"' onchange='object.changeParameter(this.id ,$(this).val())'"+
                " class='objectParam' onfocus='this.select()' size='3' value='"+obj.x+"' /></td>"+
                "<td><input type='text' id='gy"+i+"' onchange='object.changeParameter(this.id ,$(this).val())'"+
                " class='objectParam' onfocus='this.select()' size='3' value='"+obj.y+"' /></td>"+
                "<td><input type='text' id='gw"+i+"' onchange='object.changeParameter(this.id ,$(this).val())'"+
                " class='objectParam' onfocus='this.select()' size='3' value='"+obj.width+"' /></td>";
            if (obj.shape == S_CIRCLE){
                s += "<td></td><td></td>";
            } else {
                s +=
                "<td><input type='text' id='gh"+i+"' onchange='object.changeParameter(this.id ,$(this).val())'"+
                    " class='objectParam' onfocus='this.select()' size='3' value='"+obj.height+"' /></td>"+
                "<td><input type='text' id='gd"+i+"' onchange='object.changeParameter(this.id ,$(this).val())'"+
                    " class='objectParam' onfocus='this.select()' size='3' value='"+obj.degree+"' /></td>";
            }
            s += "<td><a class='tool' href='#' onmouseup='object.remove("+T_GOAL+","+i+"); revertAll(); return false;'>[D]</a></td></tr>";
            $table.append(s);
        }
    };


    this._wpListMap = function(){
        $table =  $("#tab-wps > table.container").text('');
        s = "<tr class=\"listhead\"><th>id</th><th>x</th><th>y</th>"+
            "<th>del</th></tr>";
        for (i=0; i<this.waypoint.length; i++){
            if(isEmpty(this.waypoint[i])) continue;
            s += '<tr class="object" id=agent'+ i + ' onmousedown="object.mark('+T_WAYPOINT+','+i+')" onmouseup="object.unmark('+T_WAYPOINT+','+i+')">';
            s += "<td>"+this.waypoint[i].WPID+"</td>"+
                "<td><input type='text' id='wx"+i+"' onchange='object.changeParameter(this.id,$(this).val())'"+
                (this.waypoint[i].p ? 'readonly="readonly"' : "") +
                    " class='agentParam' onfocus='this.select()' size='3' value='"+this.waypoint[i].x+"' /></td>"+
                "<td><input type='text' id='wy"+i+"' onchange='object.changeParameter(this.id,$(this).val())'"+
                (this.waypoint[i].p ? 'readonly="readonly"' : "") +
                    " class='agentParam' onfocus='this.select()' size='3' value='"+this.waypoint[i].y+"' /></td>"+
                "<td><a class='tool' onmouseup='object.remove("+T_WAYPOINT+","+i+"); revertAll(); return false;' href='#'>[D]</a></td></tr>";
        }
        $table.append(s);
    };

    this.listLink = function(){
        $table = $("#tab-linklist > table.container").text('');
        s = '<tr class="listhead"><th>from</th><th></th><th>to</th><th></th><th>del</th></tr>';
        for(i=0; i<this.links.length; i++){
            s += '<tr class="object" id="link'+i+'" onmousedown="object.markLink('+i+')"' +
                    ' onmouseup="object.unmarkLink('+i+')">'+
                '<td>' + _getIMG(this.links[i][0]) + '</td><td>' + this.links[i][2] + '</td>' +
                '<td>' + _getIMG(this.links[i][1]) + '</td><td>' + this.links[i][3] + '</td>' +
                '<td><a class="tool" href="javascript:void(0)" onmouseup="object.deleteLink('+i+'); revertAll(); return false;">[D]</a></td></tr>';
        }
        $table.append(s);
    };

    this.deleteLink = function(i){
        switch(this.links[i][0]){
        case T_START:
            switch(this.links[i][1]){
            case T_GOAL:
                this.start[this.links[i][2]].goalID = -1;
                break;
            case T_WAYPOINT:
                this.start[this.links[i][2]].firstWPID = -1;
                break;
            }
            break;
        case T_GOAL:
            this.goal[this.links[i][2]].nextStartID = -1;
            break;
        case T_WAYPOINT:
            this.waypoint[this.links[i][2]].nextWPID = -1;
            break;
        case T_PED:
            switch(this.links[i][1]){
            case T_GOAL:
                agent.agt[this.links[i][2]][2] = -1;
                break;
            case T_WAYPOINT:
                agent.agt[this.links[i][2]][3] = -1;
                break;
            }
            break;
        }
        this.links.splice(i,1);
    }
}
